
;*********************************************************************
;* Melodienklingel V 4.0											 *
;* ähnlich FA4/86; geänderte Hardware: INT-Auslöser, NMI-Betriebsart *
;* HALT+Sturmklingelsperre 											 *
;* NEU: Weihnachtslieder (über NMI) anwählbar					     *
;* Stand 08/2024													 *
;*																	 *
;* Autor Ursprungsversion: unbekannt								 *
;* Modifikationen: Günter Winkler, Falko Glanert (DG1AFG)			 *
;*																	 *
;* ACHTUNG! Es werden illegale Opcodes verwendet. 					 *
;* Daher eignen sich nicht alle Z80-CPUs beliebiger Hersteller.		 *
;* Die DDR Versionen der U880 funktionieren alle					 *
;*																	 *
;* Assembler:														 *
;* zasm -uwy GONGV40.asm -o GONGV40.bin								 *
;* https://k1.spdns.de/Develop/Projects/zasm/Distributions/			 *
;*********************************************************************

	.z80

;Grundbelegung Register 
;IY = Sprungzeiger für Noten
;IX = ACT SONG
;HL = ACT NOTE
;I  = 00H Random-Song
;   = 01H Songs der Reihe nach
;   = 02H Stop at last Song
;   = 03H Weihnachtslieder der Reihe nach
;   = 04H My Song fest
;
;Noten:
;L-Byte:Tonhoehe	'E' ist Pause
;1   2   3   4  5  6   7   8   9   A   B   C  D
;c, cis, d, dis e, f, fis, g, gis, a, ais, h, c
;einige Lieder fälschlicherweise danach codiert: ais, h, c, cis, d, dis e, f, fis, g, gis, a, ais

;H-Byte:Tonlaenge	
;BIT 4 	:Oktave --> 0=hohe Oktave, 1=niedrige Oktave
;BIT 7...5:
;000x	: 1/16
;001x	: 1/8
;010x	; 1/8 verlängert
;011x	: 1/4
;100x	; 1/4 verlängert
;101x	; 1/4 verlängert
;111x	: 1/2

NEIN	EQU	0
JA		EQU	1
EOTAB	EQU	0FFH	;Kennz. Ende Lieder-Tabelle
EOLIED	EQU	0FEH	;Kennz. Lied zu Ende
 
	ORG	0
#target Rom 

RESET:	
		DI
		IM	1
		LD 	IX,TALIED
		XOR	A		;	Random-Mode
		LD	I,A
		JR	HALT

;Sturmklingelsperre
GESTOP:				; Rücksprung vom INT
		LD	A,12	; ca 5 sec Sturmklingelsperre	
fg0:	LD	BC,0	
fg1:	DEC	C	
		JR	NZ,fg1
		DJNZ	fg1
		DEC	A
		JR	NZ,fg0

HALT:	EI			; Rücksprung vom NMI
		HALT			
		JR	HALT				   

; INT & NMI Rückkehradressen
RETI1:	DEFW	GESTOP
RETN:	DEFW	HALT

;...hier war noch Platz dafür
DELAY:	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	LD	A,(BC)
	NOP
	NOP
	JP	NEXT3

;************************************************
;*          Interruptserviceroutine             *
;*			Lied spielen						*
;************************************************
	ORG	38H

	LD	A,I			; Betriebsart holen
	JP	DOLIED

; nächste Note spielen oder fertig?
OVSKIX:
	CP	(HL)		; Liedende?
	JR	NZ,CONTX	; nein => nächste Note

	LD	SP,RETI1 	; Lied fertig
	RETI
;-------------------------------------------------

CHECK_BETRIEBSART:
;Betriebsart checken (A)
;(Verzweigung von DOLIED hierher, um den hier noch freien Speicherplatz auszunutzen)
	CP	00H
	JP	Z,DO_RANDOM	;Random

	CP	01H
	JP	Z,DO_SERIAL	;der Reihe nach

	CP	02H
	JP	Z,NXNOTE	;letzter Song

	CP	03H	
	JP	Z,DO_SERIAL	;Weihnachtslieder der Reihe nach

	LD	HL,MYSONG	
	LD	IX,MYSONG
	JP	NXNOTE		;MySong

;******************************
;*           N M I            *
;*			Betriebsart		  *
;******************************        

	ORG	66H
	
;Tastenentprellung
	LD	BC,30
nmi1:	
	DEC	C	
	JR	NZ,nmi1
	DJNZ	nmi1

;Betriebsart umschalten
	LD	A,I
	INC	A			;weiter
	CP	05H			;Maximum testen
	JR	NZ,nmi2		;...noch nicht erreicht => weiter

	XOR	A			;...erreicht => ruecksetzen
nmi2:	
	LD	I,A			;Betriebsart speichern
	CP	03H			;Betriebsart < 3 ?
	JR	C,nmi_end	;dann kein Init!

;Prüfe, welche Liedertabelle initialisiert werden soll
	LD	IX,TALIED	;init Liedertabelle Alltagslieder
	JR	NZ,nmi_end	;keine Weihnachtslieder (CP 03H weiter oben)
	LD	IX,TALIED2 	;init Weihnachtslieder

nmi_end:
	LD	SP,RETN
	RETN

; Ton erzeugen (HL)	(spielt Lied)
CONTX:	
	LD	A,1			;hohe Oktave
	BIT	4,(HL)		;welche Oktave ?
	JR	Z,$+4

	LD	A,2			;tiefe Oktave

	EX	AF,AF'		;Oktave nach A'
	LD	A,(HL)
	AND	0FH			;Tonlänge ausblenden
	LD	C,A			;Tonhöhe nach C (1-D, E=Pause)
	LD	A,(TABSIZ)	;(A:=8)
	LD	B,A			
	XOR	A
LONOTL:	
	DEC	C			;Ton testen
	JR	Z,CONT2

	ADD	A,B			;Zeiger berechnen (+8)
	JR	LONOTL		;A hat Index

CONT2:	
	DEFB	0FDH	;nicht impl. Bef. mit IY
	LD	L,A			;Ton-Zeiger nach IYL := A
	DEFB	0FDH	;nicht impl. Bef. mit IY
	LD	H,NOTEN/256 ;IYH := H-Byte NOTEN
	;=> IY hat nun den Sprungzeiger zu NOTEN
	LD	A,(HL)		;Tonbyte holen
	AND	0FH			;Tonhöhe bleibt übrig
	ADD	A,TABLE		;Umrechnung Tonhöhe
	LD	B,0			;16bit-Erw. BC
	LD	C,A
	LD	A,(BC)
	LD	E,A			;E = umgerechnete Tonhöhe
	LD	D,1			;16bit-Erw.- DE
	LD	A,(HL)		;Tonbyte noch einmal holen
;SRL r   0->7 --r--> 0->CY	
	SRL	A			;Tonlänge in das L-Nibble
	SRL	A
	SRL	A
	SRL	A
	SRL	A
;die Bits 5-7 sind nun in Bit 0-2 verschoben
	INC	A			;A= Tonlaenge
	LD	C,A			;C=    "
	EX	AF,AF'		;A= Oktave
	LD	B,A
	EX	AF,AF'
	LD	A,B			;A= Oktave
	CP	2			;tief ?
CONT3:	JR	Z,CONT4	;ja

;SLA r   CY<-7 <--r-- 0<-0
	SLA	C			;Tonlaenge *2 (weil doppelte Frequenz)
CONT4:	XOR	A		;A= Rechenregister fuer ?

LOOPOT:	ADD	A,E
	JR	NC,NOOV

	INC	D			;Uebertrag speichern
NOOV:	DEC	C
	JR	NZ,LOOPOT

LABELY:	LD	E,A
	LD	A,(HL)
	LD	C,2			;keine Pause merken
	AND	0FH			;Tonhoehe in A
	CP	0EH			;Pause ?
	JR	NZ,NEXT1 	;nein

	LD	C,0			;Pause merken 
NEXT1:	LD	A,(BC)
	LD	A,(BC)
NEXT2:	LD	A,C
	OR	A			;Pause ?
	JR	Z,NEXT3 	;ja

	OUT	(C),A		;Erzeugen des Peep (Phasenwechsel)

NEXT3:	JP	(IY)

TABSIZ:	DEFB	8	;Basis Ton-Laenge Tabelle

; Tonhöhenumrechnungstabelle
	ORG	0F0H
TABLE:	DEFB	0	;
	DEFB	3FH	;
	DEFB	42H	;
	DEFB	46H	;
	DEFB	4BH	;
	DEFB	4FH	;
	DEFB	53H	;
	DEFB	57H	;
	DEFB	5DH	;
	DEFB	63H	;
	DEFB	69H	;
	DEFB	71H	;
	DEFB	76H	;
	DEFB	7DH	;
	DEFB	3FH	;
	DEFB	0	;

	ORG	100H
NOTEN:
;c
	NOP
	LD	A,128	;
	JP	XPOINT

	ORG	NOTEN+8
;cis/des
	LD	A,I
	LD	A,120
	JP	XPOINT

	ORG	NOTEN+(2*8)
;d
	LD	SP,HL
	LD	A,113
	JP	XPOINT

	ORG	NOTEN+(3*8)
;dis/es
	LD	A,106
	NOP
	NOP
	JP	XPOINT

	ORG	NOTEN+(4*8)
;e
	LD	(TABSIZ),HL
	LD	A,99
	JP	XPOINT

	ORG	NOTEN+(5*8)
;f
	LD	A,I
	LD	A,(HL)
	LD	A,93
	JP	XPOINT

	ORG	NOTEN+(6*8)
;fis/ges
	NOP
	LD	A,88
	JP	XPOINT

	ORG	NOTEN+(7*8)
;g
	LD	A,(TABSIZ)
	LD	A,82
	JP	XPOINT

	ORG	NOTEN+(8*8)
;gis/as
	LD	A,(HL)
	NOP
	LD	A,77
	JP	XPOINT

	ORG	NOTEN+(9*8)
;a
	LD	A,(TABSIZ)
	LD	A,72
	JP	XPOINT

	ORG	NOTEN+(10*8)
;ais/b
	LD	SP,HL
	LD	A,68
	JP	XPOINT

	ORG	NOTEN+(11*8)
;h
	LD	A,I
	LD	A,(HL)
	LD	A,63
	JP	XPOINT

	ORG	NOTEN+(12*8)
;c
	LD	A,I
	LD	A,(HL)
	LD	A,59
	JP	XPOINT

	ORG	NOTEN+(13*8)
;pause
	NOP
	LD	A,129
XPOINT:	DEC	A	;Verzoegerung 16tel
	JR	NZ,XPOINT

	DEC	B
	JP	NZ,DELAY

	EX	AF,AF'
	LD	B,A
	EX	AF,AF'
	LD	A,B	;Oktave auf A' geholt
	DEC	E
	JP	NZ,NEXT1

	DEC	D	
	JP	NZ,NEXT2

	JR	DELAY2	;fertig


DELAY2:	LD	DE,4	;Nachverzoegerung
LOODEL:	DEC	D
		JR	NZ,LOODEL

		DEC	E
		JR	NZ,LOODEL

		INC	HL	;naechster Ton
NXNOTE:	
;	(naechste) Note spielen
		LD	A,EOLIED 
		JP	OVSKIX


;------------------------------------------
DOLIED:
;--verbotener Befehl: DE:=IX
	DEFB	0DDH 	;nicht impl. IX: LD E, IXL
	LD	E,L
	DEFB	0DDH 	;nicht impl. IX: LD D, IXH
	LD	D,H			; => DE := IX
	EX	DE,HL		; => HL := IX
;--
	JP	CHECK_BETRIEBSART	; (zwischen INT-ISR und NMI war halt noch freier Platz...)

DO_SERIAL:
	LD	A,1		;DJNZ (SUCH) nicht ausführen
	JR	Nx_SUCH	;=> Songs der Reihe nach

DO_RANDOM:
	LD	A,R		;Generierung Zufallswert (7 Bit Refresh)
;	RRCA		;:2
	AND	7FH		;in Tabelle max. 128 weiter

Nx_SUCH:	
	LD	B,A		;Reg B fuer DJNZ initialisieren

SUCH:	
	LD	A,EOLIED
SUCH_LIEDENDE:
	CP	(HL)
	INC	HL
	JR	NZ,SUCH_LIEDENDE
;Liedende jetzt erreicht 
	LD	A,EOTAB 	
	CP	(HL)		;Test auf Ende der Tabelle
	JR	NZ,SUCH_WEITER

;Tabellenende erreicht & Neuinitialisierung
	LD	HL,TALIED	;init Alltagslieder
	LD	A,I			;Betriebsart
	CP	03H			;Weihnachten?
	JR	NZ,SUCH_WEITER ;nein

	LD	HL,TALIED2	;Weihnachtslieder
SUCH_WEITER:
	DJNZ	SUCH	;max 16 Lieder weiter bei Random

	EX	DE,HL		;jetzt hat er das neue Lied...
	LD	IX,0
	ADD	IX,DE		;IX := HL 
	EX	DE,HL		;Liedanfang muss in IX und HL stehen !
	JP	NXNOTE		;Lied spielen
;
	DEFB	EOLIED

;Liedertabellen einbinden
	include "./LIEDER.inc"

;FIN	EQU	$


	.end
